#include "stdafx.h"
//////////////////////////////////////////////////////////////////////////////////////////////////
// 
// File: camera.cpp
// 
// Author: Frank Luna (C) All Rights Reserved
//
// System: AMD Athlon 1800+ XP, 512 DDR, Geforce 3, Windows XP, MSVC++ 7.0 
//
// Desc: Defines a camera's position and orientation.
//         
//////////////////////////////////////////////////////////////////////////////////////////////////
#include "camera.h"

Camera::Camera()
{
	_cameraType = AIRCRAFT;
	
	// Set for projection matrix.
	//    
	_fovAngle = D3DX_PI * 0.5f; 
	_aspect	  = (float)800 / (float)600;
	_zNear	  = 1.0f;
	_zFar	  = 1000.0f;

	_pos   = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
	_right = D3DXVECTOR3(1.0f, 0.0f, 0.0f);
	_up    = D3DXVECTOR3(0.0f, 1.0f, 0.0f);
	_look  = D3DXVECTOR3(0.0f, 0.0f, 1.0f);
}

Camera::Camera(CameraType cameraType)
{
	_cameraType = cameraType;
	
	// Set for projection matrix.
	//    
	_fovAngle = D3DX_PI * 0.5f; 
	_aspect	  = (float)800 / (float)600;
	_zNear	  = 1.0f;
	_zFar	  = 1000.0f;

	_pos   = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
	_right = D3DXVECTOR3(1.0f, 0.0f, 0.0f);
	_up    = D3DXVECTOR3(0.0f, 1.0f, 0.0f);
	_look  = D3DXVECTOR3(0.0f, 0.0f, 1.0f);
}

Camera::~Camera()
{

}

void Camera::getPosition(D3DXVECTOR3* pos)
{
	*pos = _pos;
}

void Camera::setPosition(D3DXVECTOR3* pos)
{
	_pos = *pos;
}

void Camera::getRight(D3DXVECTOR3* right)
{
	*right = _right;
}

void Camera::getUp(D3DXVECTOR3* up)
{
	*up = _up;
}

void Camera::getLook(D3DXVECTOR3* look)
{
	*look = _look;
}

void Camera::setRight(D3DXVECTOR3* right)
{
	_right = *right;
}

void Camera::setUp(D3DXVECTOR3* up)
{
	_up = *up;
}

void Camera::setLook(D3DXVECTOR3* look)
{
	_look = *look;
}

void Camera::walk(float units)
{
	// move only on xz plane for land object
	if( _cameraType == LANDOBJECT )
		_pos += D3DXVECTOR3(_look.x, 0.0f, _look.z) * units;

	if( _cameraType == AIRCRAFT )
		_pos += _look * units;
}

void Camera::strafe(float units)
{
	// move only on xz plane for land object
	if( _cameraType == LANDOBJECT )
		_pos += D3DXVECTOR3(_right.x, 0.0f, _right.z) * units;

	if( _cameraType == AIRCRAFT )
		_pos += _right * units;
}

void Camera::fly(float units)
{
	// move only on y-axis for land object
	if( _cameraType == LANDOBJECT )
		_pos.y += units;

	if( _cameraType == AIRCRAFT )
		_pos += _up * units;
}

void Camera::pitch(float angle)
{
	D3DXMATRIX T;
	D3DXMatrixRotationAxis(&T, &_right,	angle);

	// rotate _up and _look around _right vector
	D3DXVec3TransformCoord(&_up,&_up, &T);
	D3DXVec3TransformCoord(&_look,&_look, &T);
}

void Camera::yaw(float angle)
{
	D3DXMATRIX T;

	// rotate around world y (0, 1, 0) always for land object
	if( _cameraType == LANDOBJECT )
		D3DXMatrixRotationY(&T, angle);

	// rotate around own up vector for aircraft
	if( _cameraType == AIRCRAFT )
		D3DXMatrixRotationAxis(&T, &_up, angle);

	// rotate _right and _look around _up or y-axis
	D3DXVec3TransformCoord(&_right,&_right, &T);
	D3DXVec3TransformCoord(&_look,&_look, &T);
}

void Camera::roll(float angle)
{
	// only roll for aircraft type
	if( _cameraType == AIRCRAFT )
	{
		D3DXMATRIX T;
		D3DXMatrixRotationAxis(&T, &_look,	angle);

		// rotate _up and _right around _look vector
		D3DXVec3TransformCoord(&_right,&_right, &T);
		D3DXVec3TransformCoord(&_up,&_up, &T);
	}
}

void Camera::getViewMatrix(D3DXMATRIX* V)
{
	// Keep camera's axes orthogonal to eachother
	D3DXVec3Normalize(&_look, &_look);

	D3DXVec3Cross(&_up, &_look, &_right);
	D3DXVec3Normalize(&_up, &_up);

	D3DXVec3Cross(&_right, &_up, &_look);
	D3DXVec3Normalize(&_right, &_right);

	// Build the view matrix:
	float x = -D3DXVec3Dot(&_right, &_pos);
	float y = -D3DXVec3Dot(&_up, &_pos);
	float z = -D3DXVec3Dot(&_look, &_pos);

	(*V)(0,0) = _right.x; (*V)(0, 1) = _up.x; (*V)(0, 2) = _look.x; (*V)(0, 3) = 0.0f;
	(*V)(1,0) = _right.y; (*V)(1, 1) = _up.y; (*V)(1, 2) = _look.y; (*V)(1, 3) = 0.0f;
	(*V)(2,0) = _right.z; (*V)(2, 1) = _up.z; (*V)(2, 2) = _look.z; (*V)(2, 3) = 0.0f;
	(*V)(3,0) = x;        (*V)(3, 1) = y;     (*V)(3, 2) = z;       (*V)(3, 3) = 1.0f;
}

void Camera::getProjMatrix(D3DXMATRIX* Proj)
{
	Proj = &_Proj;
}

void Camera::getViewProjMatrix(D3DXMATRIX* ViewProj)
{
	D3DXMATRIX view,proj;
	Camera::getViewMatrix(&view);
	Camera::getProjMatrix(&proj);
	(*ViewProj) = view * proj;
}

void Camera::getSkyBoxMatrix(D3DXMATRIX* SkyBox, IDirect3DDevice9* device)
{
	//       
	//    
	D3DXMATRIX w, vm;
	Camera::getViewMatrix(&vm);
//	device->GetTransform(D3DTS_WORLD, &w);
//	float determinant;	
//	D3DXMatrixInverse(SkyBox, &determinant, &w);

	(*SkyBox) = vm; //  ,   

	(*SkyBox)._41 = 0.0f;
	(*SkyBox)._42 = 0.0f;
	(*SkyBox)._43 = 0.0f;

	//      ,   
	//    
	
	D3DXMatrixMultiply(
		SkyBox, 
		SkyBox, 
		&_matUnitProj);	
	
}

void Camera::ReadInput( float timeDelta )
{
	//  --------------------------------
		static float megaspeed = 1.0f;
		//  	
		// elapsedTime - - ,    startTime 
		static DWORD startTime2 = ::GetTickCount(); 
		DWORD elapsedTime2 = ::GetTickCount() - startTime2;	
		if(elapsedTime2>10)
		{	
			POINT CurPosNow;
			GetCursorPos(&CurPosNow);
			SetCursor( NULL );
			pitch( (CurPosNow.y - 300) * timeDelta * 0.1 );
			yaw( (CurPosNow.x - 400) * timeDelta * 0.1 );
			SetCursorPos(400, 300);
			startTime2 = ::GetTickCount();
		}
		if( ::GetAsyncKeyState('W') & 0x8000f )
			walk(4.0f * timeDelta * megaspeed);

		if( ::GetAsyncKeyState('S') & 0x8000f )
			walk(-4.0f * timeDelta*  megaspeed);

		if( ::GetAsyncKeyState('A') & 0x8000f )
			strafe(-4.0f * timeDelta * megaspeed);

		if( ::GetAsyncKeyState('D') & 0x8000f )
			strafe(4.0f * timeDelta * megaspeed);

		if( ::GetAsyncKeyState(VK_SPACE) & 0x8000f )
			fly(4.0f * timeDelta * megaspeed);

		if( ::GetAsyncKeyState(VK_LCONTROL) & 0x8000f )
			fly(-4.0f * timeDelta * megaspeed);

		if( ::GetAsyncKeyState(VK_UP) & 0x8000f )
			pitch(1.0f * timeDelta * megaspeed / 4);

		if( ::GetAsyncKeyState(VK_DOWN) & 0x8000f )
			pitch(-1.0f * timeDelta * megaspeed / 4);

		if( ::GetAsyncKeyState(VK_LEFT) & 0x8000f )
			yaw(-1.0f * timeDelta * megaspeed / 4);
			
		if( ::GetAsyncKeyState(VK_RIGHT) & 0x8000f )	
			yaw(1.0f * timeDelta * megaspeed / 4);	

		if( ::GetAsyncKeyState('N') & 0x8000f )
			roll(1.0f * timeDelta * megaspeed);

		if( ::GetAsyncKeyState('M') & 0x8000f )
			roll(-1.0f * timeDelta * megaspeed);

		if( ::GetAsyncKeyState(VK_LSHIFT) & 0x8000f )
			 megaspeed = 18.0f;
		else megaspeed = 1.0f;
}

void Camera::useCamera( IDirect3DDevice9*  device )
{
	D3DXMATRIX VM;
	getViewMatrix( &VM );
	device->SetTransform(D3DTS_VIEW, &VM);
}

void Camera::setProjParam(IDirect3DDevice9*  device, float fov, float asp, float zN, float zF)
{
	if( fov )	_fovAngle = fov;
	if( asp )   _aspect = asp;
	if( zN )	_zNear = zN;
	if( zF )	_zFar =zF;	

	D3DXMatrixPerspectiveFovLH(
			&_Proj,
			_fovAngle, 
			_aspect,
			_zNear,
			_zFar);
	
	//       
	// (      )
	D3DXMatrixPerspectiveFovLH(
			&_matUnitProj,
			_fovAngle, 
			_aspect,
			0.01f,
			2.0f);
	

	device->SetTransform(D3DTS_PROJECTION, &_Proj);
}

void Camera::focusOn( D3DXVECTOR3* point, float speed )
{
	//    XZ
	float Ugolxz = UgTurnVec2D(_pos.x, _pos.z, 
								    _look.x, _look.z, 
								    (*point).x, (*point).z);
	//  ZY
	float Ugolzy = UgTurnVec2D(_pos.z, _pos.y, 
								    _look.z, _look.y, 
								    (*point).z, (*point).y);
	//  _look   Z
	yaw(Ugolxz * speed);
	//  _look   ZY
	// ,  _look    
	float t_ug = Ugol2D(_pos.z, _pos.y, (*point).z, (*point).y);
	if ( (t_ug > D3DX_PI / 2)&&(t_ug < 3*D3DX_PI/2) ) Ugolzy = -Ugolzy;
	//  = 90  270  ,   
	if( (t_ug == D3DX_PI / 2)||(t_ug == 3*D3DX_PI / 2) ) ;
	else pitch(Ugolzy * speed);

}

void Camera::setCameraType(CameraType cameraType)
{
	_cameraType = cameraType;
}

float Camera::Ugol2D(float X1, float Y1, float X2, float Y2)
{
	//   
	float x1 = 0.0f, y1 = 0.0f;
	float x2 = X2 - X1;
	float y2 = Y2 - Y1;
	if(x2 == 0)
	{
		if (y2 < 0) return D3DX_PI/2;
		else if(y2 > 0) return D3DX_PI * 3 / 2;
		else return 0.0f;
	}
	else
	{
		if (y2 == 0)
		{
			if (x2>=0)  return 0.0f;
			else return D3DX_PI;
		}
		// 
		if ((x2>0)&&(y2>0)) return 2*D3DX_PI - abs( atanf(y2/x2) );
		if ((x2<0)&&(y2>0)) return D3DX_PI + abs( atanf(y2/x2) );
		if ((x2<0)&&(y2<0)) return D3DX_PI - abs( atanf(y2/x2) );
		if ((x2>0)&&(y2<0)) return abs( atanf(y2/x2) );
	}
	return 0.0f;
}

float Camera::UgTurnVec2D(float Xr, float Yr, float dx, float dy, float Xt, float Yt)
{
	float Maxa = 0.0f;
	float ugb = Ugol2D(Xr, Yr, Xt, Yt);
	float uga = Ugol2D(Xr, Yr, Xr + dx, Yr + dy);
	if (abs(uga-ugb) > D3DX_PI /360) //    1 
	{
		if (uga<ugb) Maxa = 1.0f;
		if (uga>ugb) Maxa = -1.0f;
		if(abs(uga-ugb) >= D3DX_PI)
		{	//  ,     maxa
		/*	if ( abs(uga-ugb) >= maxa) uga -= Maxa;
			else uga -=  abs(uga-ugb);	*/
			return -1.0f*Maxa;
		}
		else
		{
		/*	if ( abs(uga-ugb)>=maxa ) uga += Maxa;
			else uga += abs(uga-ugb);	*/
			return 1.0f*Maxa;
		}
	}
	else return 0.0f;
}